# -*- coding: utf-8 -*-

from imio.schedule.content.task import IAutomatedTask
from imio.schedule.content.task_config import ITaskConfig

from plone import api

from zope.component import getUtility
from zope.component.interface import getInterface
from zope.interface import alsoProvides
from zope.interface import noLongerProvides
from zope.schema.interfaces import IVocabularyFactory


def update_marker_interfaces(task_config, event):
    """
    When the 'marker_interfaces' field is updated, update the interface
    provided on all the tasks of this task config as well.
    """
    catalog = api.portal.get_tool('portal_catalog')

    vocname = ITaskConfig.get('marker_interfaces').value_type.vocabularyName
    interfaces_voc = getUtility(IVocabularyFactory, vocname)(task_config)
    marker_interfaces = dict([(i, getInterface('', i)) for i in interfaces_voc.by_value])

    task_brains = catalog(
        object_provides=IAutomatedTask.__identifier__,
        task_config_UID=task_config.UID()
    )
    sample_task = task_brains and task_brains[0].getObject() or None

    # verify if the update is needed
    do_update = False
    for interface_name, marker_interface in marker_interfaces.iteritems():
        is_provided = marker_interface.providedBy(sample_task)
        # new interface on the config but not present yet on the tasks => update
        if interface_name in (task_config.marker_interfaces or []) and not is_provided:
            do_update = True
            break
        # old interface on the tasks no longer present on the config => update
        elif interface_name not in (task_config.marker_interfaces or []) and is_provided:
            do_update = True
            break

    if do_update:
        for task_brain in task_brains:
            task = task_brain.getObject()

            for marker_interface_name, marker_interface in marker_interfaces.iteritems():
                if marker_interface_name in task_config.marker_interfaces:
                    alsoProvides(task, marker_interface)
                else:
                    noLongerProvides(task, marker_interface)

            task.reindexObject(idxs=['object_provides'])



# -*- coding: utf-8 -*-

from collective.wfadaptations.interfaces import IWorkflowAdaptation

from Persistence import PersistentMapping

from plone import api

from zope.interface import implements


class FreezeTaskWorkflowAdaptation(object):
    """
    Suspension workflow adaptation adding a suspension state.
    """

    implements(IWorkflowAdaptation)

    schema = None

    def patch_workflow(self, workflow_name, **parameters):

        wtool = api.portal.get_tool('portal_workflow')
        workflow = wtool[workflow_name]

        self.create_frozen_state(workflow, **parameters)

        message = "patched '{}' workflow with frozen state".format(workflow_name)
        return True, message

    def create_frozen_state(self, workflow, **parameters):
        """
        create a 'frozen' state
        """
        if 'frozen' not in workflow.states:
            workflow.states.addState('frozen')

        frozen_state = workflow.states['frozen']
        default_mapping = workflow.states.objectValues()[0].permission_roles.copy()
        frozen_state.title = 'frozen'
        frozen_state.permission_roles = default_mapping
        frozen_state.group_roles = PersistentMapping()
        frozen_state.var_values = PersistentMapping()
        frozen_state.transitions = ()



# -*- coding: utf-8 -*-

from Acquisition import aq_base

from plone import api

from imio.schedule.interfaces import TaskAlreadyExists
from imio.schedule.utils import get_task_configs
from imio.schedule.utils import get_container_open_tasks


class TaskEventHandler(object):

    def __init__(self, task_container, event):
        self.container = task_container
        self.event = event

        if self.is_created is False:
            return

        self.task_configs = get_task_configs(self.container)
        if not self.task_configs:
            return

        with api.env.adopt_roles(['Manager']):
            self.handle()

    @property
    def is_created(self):
        """
        This event can be triggered for archetype containers after a workflow
        modification so we need to ensured that the container really exists
        (more than just created in portal_factory).
        """
        if hasattr(aq_base(self.container), 'checkCreationFlag'):
            if self.container.checkCreationFlag():
                return False
        return True

    def handle(self):
        """
        """
        raise NotImplementedError


def start_tasks(task_container, event):
    """
    Automatically start tasks matching start conditions of a given
    task container.
    """

    # This handler can be triggered for archetypes containers by the
    # workflow modification event but we want to create tasks only if
    # the container really exists (more than just created in portal_factory...)
    if hasattr(aq_base(task_container), 'checkCreationFlag'):
        if task_container.checkCreationFlag():
            return

    task_configs = get_task_configs(task_container)

    if not task_configs:
        return

    with api.env.adopt_roles(['Manager']):
        for config in task_configs:
            task = config.get_created_task(task_container)
            if task and config.should_start_task(task_container, task):
                # delegate the starting action to the config so different behaviors
                # can be easily configured
                config.start_task(task)

                # update due date after the task has started, because some date
                # computation can rely on the task starting date
                task.due_date = config.compute_due_date(task_container, task)
                task.reindexObject(idxs=('due_date',))


def end_tasks(task_container, event):
    """
    Automatically end tasks matching end conditions of a given
    task container.
    """

    # This handler can be triggered for archetypes containers by the
    # workflow modification event but we want to create tasks only if
    # the container really exists (more than just created in portal_factory...)
    if hasattr(aq_base(task_container), 'checkCreationFlag'):
        if task_container.checkCreationFlag():
            return

    task_configs = get_task_configs(task_container)

    if not task_configs:
        return

    with api.env.adopt_roles(['Manager']):
        for config in task_configs:
            task = config.get_open_task(task_container)
            if task and config.should_end_task(task_container, task):
                # delegate the closure action to the config so different behaviors
                # can be easily configured
                config.end_task(task)


def freeze_tasks(task_container, event):
    """
    Automatically freeze tasks matching freeze conditions of a given
    task container.
    """

    # This handler can be triggered for archetypes containers by the
    # workflow modification event but we want to create tasks only if
    # the container really exists (more than just created in portal_factory...)
    if hasattr(aq_base(task_container), 'checkCreationFlag'):
        if task_container.checkCreationFlag():
            return

    task_configs = get_task_configs(task_container)

    if not task_configs:
        return

    with api.env.adopt_roles(['Manager']):
        for config in task_configs:
            task = config.get_open_task(task_container)
            if task and config.should_freeze_task(task_container, task):
                # delegate the freeze action to the config so different behaviors
                # can be easily configured
                config.freeze_task(task)

                # update due date after the task has frozen
                task.due_date = config.compute_due_date(task_container, task)
                task.reindexObject(idxs=('due_date',))


def thaw_tasks(task_container, event):
    """
    Automatically thaw tasks matching thaw conditions of a given
    task container.
    """

    # This handler can be triggered for archetypes containers by the
    # workflow modification event but we want to create tasks only if
    # the container really exists (more than just created in portal_factory...)
    if hasattr(aq_base(task_container), 'checkCreationFlag'):
        if task_container.checkCreationFlag():
            return

    task_configs = get_task_configs(task_container)

    if not task_configs:
        return

    with api.env.adopt_roles(['Manager']):
        for config in task_configs:
            task = config.get_frozen_task(task_container)
            if task and config.should_thaw_task(task_container, task):
                # delegate the thaw action to the config so different behaviors
                # can be easily configured
                config.thaw_task(task)

                # update due date after the task has thawed, because some date
                # computation can rely on the task thawing date
                task.due_date = config.compute_due_date(task_container, task)
                task.reindexObject(idxs=('due_date',))


def update_due_date(task_container, event):
    """
    If the task_container has been modified, compute
    the due date again and update it on the task.
    """

    # This handler can be triggered for archetypes containers by the
    # workflow modification event but we want to create tasks only if
    # the container really exists (more than just created in portal_factory...)
    if hasattr(aq_base(task_container), 'checkCreationFlag'):
        if task_container.checkCreationFlag():
            return

    task_configs = get_task_configs(task_container)

    if not task_configs:
        return

    with api.env.adopt_roles(['Manager']):
        for config in task_configs:
            task = config.get_open_task(task_container)
            if task:
                task.due_date = config.compute_due_date(task_container, task)
                task.reindexObject(idxs=('due_date',))


class UpdateRecurrenceHandler(TaskEventHandler):

    def handle(self):
        open_tasks = get_container_open_tasks(self.container)
        open_tasks_by_cfg_UIDs = dict([(t.task_config_UID, t) for t in open_tasks])
        for config in self.task_configs:
            if config.is_main_taskconfig():
                has_open_task = config.UID() in open_tasks_by_cfg_UIDs.keys()
                if not has_open_task and config.should_recurred(self.container):
                    try:
                        config.create_recurring_task(self.container)
                    except TaskAlreadyExists:
                        continue
            # case of a sub-task creation, the parent should have been created first
            else:
                macro_config = config.getParentNode()
                has_parent_task = macro_config.UID() in open_tasks_by_cfg_UIDs.keys()
                if has_parent_task and config.should_recurred(self.container):
                    parent_task = open_tasks_by_cfg_UIDs[macro_config.UID()]
                    if config.get_open_task(parent_task) is not None:
                        continue
                    config.create_recurring_task(
                        self.container,
                        creation_place=parent_task,
                    )



# -*- coding: utf-8 -*-

from zope.interface import Interface
from zope import schema
from z3c.form.form import Form
from z3c.form import button
from z3c.form import field
from Products.statusmessages.interfaces import IStatusMessage
from plone.z3cform.layout import FormWrapper
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile

from imio.schedule import _


class ITaskChangeOwner(Interface):

    new_owner = schema.Choice(
        title=_(u'New owner'),
        vocabulary='schedule.task_owner',
        required=True,
    )


class TaskChangeOwnerForm(Form):
    fields = field.Fields(ITaskChangeOwner)
    ignoreContext = True

    @button.buttonAndHandler(_(u'Confirm'))
    def handleApply(self, action):
        messages = IStatusMessage(self.request)
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            messages.addStatusMessage(self.status, type="error")
            return
        self.context.assigned_user = data.get('new_owner')
        self.context.reindexObject()
        self.context.reindex_parent_tasks(idxs=['is_solvable_task'])
        self.status = _(u'Owner changed')


class TaskChangeOwnerView(FormWrapper):
    form = TaskChangeOwnerForm
    index = ViewPageTemplateFile('templates/task_change_owner.pt')



# -*- coding: utf-8 -*-

from datetime import datetime
from mock import Mock
from DateTime import DateTime

from imio.schedule.content.delay import CalculationDefaultDelay
from imio.schedule.testing import ExampleScheduleFunctionalTestCase
from imio.schedule.tests.due_date import ContainerCreationDate


class TestCalculationDelay(ExampleScheduleFunctionalTestCase):

    def setUp(self):
        super(TestCalculationDelay, self).setUp()
        self._start_date = ContainerCreationDate.start_date

    def tearDown(self):
        ContainerCreationDate.start_date = self._start_date
        super(TestCalculationDelay, self).tearDown()

    @property
    def delay_adapter(self):
        return CalculationDefaultDelay(self.task_container, self.task)

    def test_start_date(self):
        """Test different cases for 'start_date' property"""
        current_date = datetime.now().date()
        self.assertEqual(current_date, self.delay_adapter.start_date)

        ContainerCreationDate.start_date = Mock(return_value=DateTime())
        self.assertEqual(current_date, self.delay_adapter.start_date)

        ContainerCreationDate.start_date = Mock(return_value=None)
        self.assertEqual(
            datetime(9999, 1, 1).date(),
            self.delay_adapter.start_date,
        )

    def test_delta(self):
        """Test 'delta' property"""
        adapter = self.delay_adapter

        adapter.calculate_delay = Mock(return_value=0)
        self.assertEqual(0, adapter.delta)

        adapter.calculate_delay = Mock(return_value=10)
        self.assertEqual(10, adapter.delta)

    def test_due_date(self):
        """Test 'due_date' property"""
        adapter = self.delay_adapter
        ContainerCreationDate.start_date = Mock(return_value=datetime(2016, 1, 1).date())
        due_date = datetime(2016, 1, 11).date()
        adapter.calculate_delay = Mock(return_value=10)

        self.assertEqual(due_date, adapter.due_date)

    def test_compute_due_date(self):
        """Test 'compute_due_date' method"""
        adapter = self.delay_adapter
        base_date = datetime(2016, 1, 1).date()
        due_date = datetime(2016, 1, 11).date()
        adapter.calculate_delay = Mock(return_value=10)

        self.assertEqual(due_date, adapter.compute_due_date(base_date))



# encoding: utf-8

from imio.schedule.config import DONE
from imio.schedule.config import STARTED
from imio.schedule.config import status_by_state

from plone import api

from Products.Five import BrowserView


class TaskCompletionView(BrowserView):
    """
    View of the popup showing the completion details of a task.
    Display the status of each start/end condition of the task.
    Display if the starting/ending state is matched or not.
    """

    subtask_title_label = 'Title'
    subtask_todo_title_label = 'Title'
    subtask_status_label = 'Status'
    due_date_label = 'Due date'
    end_date_label = 'End date'
    assigned_user_label = 'Assigned to'

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.task = context

    def get_conditions_status(self):
        """
        List all the tasks conditions status (except for workflow state).
        """

    def get_state_status(self):
        """
        """

    def get_subtasks_status(self):
        """
        """
        created = []
        started = []
        done = []
        for subtask in self.task.get_last_subtasks():
            status = status_by_state[self.get_state(subtask)]
            if status is DONE:
                done.append(subtask)
            elif status is STARTED:
                started.append(subtask)
            else:
                created.append(subtask)

        return created, started, done

    def get_state(self, context):
        """
        Return the context workflow state.
        """
        return api.content.get_state(context)

    def display_user_name(self, user_id):
        """
        Return the full name of the given user id.
        """
        user = api.user.get(user_id)
        user_name = user and user.getProperty('fullname') or user_id
        return user_name

    def display_date(self, date):
        """
        """
        if not date:
            return '-'
        if date.year == 9999:
            return u'\u221E'.encode('utf-8')

        return date.strftime('%d/%m/%Y')


class TaskStartStatusView(TaskCompletionView):
    """
    View of the popup showing the start completion details of a created task.
    Display the status of each start condition of the task.
    Display the status of each subtask.
    Display if the starting state is matched or not.
    """

    def get_conditions_status(self):
        """
        List all the tasks conditions status (except for workflow state).
        """
        return self.task.start_conditions_status()

    def get_state_status(self):
        """
        """
        return self.task.starting_states_status()


class TaskStartSimpleStatusView(TaskStartStatusView):
    """
    Same as the above but without subtasks status.
    """

    def get_subtasks_status(self):
        """
        Do not display subtasks.
        """
        return [], [], []


class TaskEndStatusView(TaskCompletionView):
    """
    View of the popup showing the end completion details of a started task.
    Display the status of each end condition of the task.
    Display if the ending state is matched or not.
    """

    def get_conditions_status(self):
        """
        List all the tasks conditions status (except for workflow state).
        """
        return self.task.end_conditions_status()

    def get_state_status(self):
        """
        """
        return self.task.ending_states_status()


class TaskEndSimpleStatusView(TaskEndStatusView):
    """
    Same as the above but without subtasks status.
    """

    def get_subtasks_status(self):
        """
        Do not display subtasks.
        """
        return [], [], []



# -*- coding: utf-8 -*-

from datetime import date
from datetime import datetime
from dateutil.relativedelta import relativedelta
from zope.annotation import IAnnotations
from zope.component import queryMultiAdapter

from imio.schedule.content.logic import TaskLogic
from imio.schedule.interfaces import IStartDate


class BaseCalculationDelay(TaskLogic):

    start_date_interface = IStartDate

    @property
    def start_date(self):
        """Return the start date for the task container and the task"""
        date_adapter = queryMultiAdapter(
            (self.task_container, self.task),
            interface=self.start_date_interface,
            name=self.task_config.start_date,
        )
        start_date = date_adapter and date_adapter.start_date()
        if not start_date:
            start_date = date(9999, 1, 1)
        if not isinstance(start_date, date):
            start_date = start_date.asdatetime().date()
        return start_date

    @property
    def delta(self):
        """Return the calculated delay + the extra delay"""
        return self.calculate_delay()

    @property
    def due_date(self):
        """Return the due date calculated with the delay"""
        return self.start_date + relativedelta(days=+self.delta)

    def calculate_delay(self):
        """Method that must be overrided to calculate the delay"""
        raise NotImplementedError

    def compute_due_date(self, date):
        """Return the addition of the given date and the extra delay"""
        return date + relativedelta(days=+self.delta)


class CalculationDefaultDelay(BaseCalculationDelay):
    """
    """

    def calculate_delay(self):
        return 0


class DefaultFreezeDuration(object):
    """
    """

    def __init__(self, task_container, task):
        self.container = task_container
        self.task = task

    @property
    def freeze_duration(self):
        annotations = IAnnotations(self.task)
        freeze_infos = annotations['imio.schedule.freeze_task']
        freeze_date = datetime.strptime(freeze_infos['freeze_date'], '%Y-%m-%d')
        freeze_delta = datetime.now().date() - freeze_date.date()
        new_freeze_duration = freeze_infos['previous_freeze_duration'] + freeze_delta.days
        return new_freeze_duration



# -*- coding: utf-8 -*-

from datetime import date
from zope.interface import implements

from imio.schedule.interfaces import ICalendarExtraHolidays


class CalendarExtraHolidays(object):
    implements(ICalendarExtraHolidays)

    def get_holidays(self, year):
        return ((date(2017, 2, 1), 'Test holidays'), )



# -*-coding: utf-8 -*-

from plone import api
from plone.dexterity.content import Container
from plone.supermodel import model

from imio.schedule import _
from imio.schedule.content.task_config import ITaskConfig
from imio.schedule.utils import tuple_to_interface

from plone.memoize.request import cache

from zope import schema
from zope.interface import implements


class IScheduleConfig(model.Schema):
    """
    ScheduleConfig dexterity schema.
    """

    enabled = schema.Bool(
        title=_(u'Enabled'),
        default=True,
        required=False,
    )

    scheduled_contenttype = schema.Choice(
        title=_(u'Scheduled content type'),
        description=_(u'Select the content type to apply schedule.'),
        vocabulary='schedule.scheduled_contenttype',
        required=True,
    )


class ScheduleConfig(Container):
    """
    ScheduleConfig dexterity class.
    """

    implements(IScheduleConfig)

    def level(self):
        """
        Return depth contenance level.
        """
        return 0

    def query_task_configs(self, query={}):
        """
        Query the TaskConfig of this ScheduleConfig.
        """
        catalog = api.portal.get_tool('portal_catalog')
        config_path = '/'.join(self.getPhysicalPath())

        base_query = {
            'object_provides': ITaskConfig.__identifier__,
            'path': {'query': config_path},
            'sort_on': 'getObjPositionInParent',
        }

        base_query.update(query)

        config_brains = catalog(**base_query)

        return config_brains

    @cache(get_key=lambda m, schedule_cfg: schedule_cfg.id, get_request='self.REQUEST')
    def get_all_task_configs(self):
        """
        Return all the TaskConfig of this ScheduleConfig.
        """
        config_brains = self.query_task_configs()
        with api.env.adopt_roles(['Manager']):
            task_configs = [brain.getObject() for brain in config_brains]

        return task_configs

    def get_scheduled_portal_type(self):
        """
        Return the portal_type of the selected scheduled_contenttype.
        """
        return self.scheduled_contenttype and self.scheduled_contenttype[0] or ''

    def get_scheduled_interfaces(self):
        """
        Return the registration interface of the selected scheduled_contenttype.
        """
        portal_type, interface_tuples = self.scheduled_contenttype
        if type(interface_tuples[0]) not in [list, tuple]:
            interface_tuples = (interface_tuples,)
        interfaces = tuple([tuple_to_interface(i) for i in interface_tuples])

        return interfaces

    def is_empty(self):
        """
        Tells if the schedule config has no task configs.
        """
        return len(self.objectIds()) <= 1



# -*- coding: utf-8 -*-

from collective.eeafaceted.collectionwidget.browser.views import RenderTermView

from imio.schedule.interfaces import IToIcon

from zope.component import queryAdapter


class RenderScheduleTermView(RenderTermView):
    """
    """

    selected_term = ''

    def get_icon(self):
        to_icon = queryAdapter(self.context, IToIcon)
        if to_icon:
            return to_icon.get_icon_url()
        return ''



# -*- coding: utf-8 -*-

from imio.schedule.config import task_types

from plone import api

from zope.i18n import translate as _
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary


class TaskWorkflowStates(object):
    """
    List all states of urban licence workflow.
    """

    def __call__(self, context):

        states = ['created', 'to_do', 'closed']

        vocabulary_terms = []
        for state in states:
            vocabulary_terms.append(
                SimpleTerm(
                    state,
                    state,
                    _(state, 'collective.task', context.REQUEST)
                )
            )

        vocabulary = SimpleVocabulary(vocabulary_terms)
        return vocabulary


class TaskPortalTypes(object):
    """
    List all states of urban licence workflow.
    """

    def __call__(self, context):

        portal_types = api.portal.get_tool('portal_types')
        vocabulary_terms = []

        for task_type in task_types:
            p_type = portal_types[task_type]
            vocabulary_terms.append(
                SimpleTerm(
                    task_type,
                    task_type,
                    _(p_type.Title(), 'imio.schedule', context.REQUEST)
                )
            )

        vocabulary = SimpleVocabulary(vocabulary_terms)
        return vocabulary
